/*******************************************************************************
 * Copyright (c) 2006 Cisco Technology, Inc and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Author: Andrew Cheadle, CrossCore Optimization Limited
 *
 * Contributors:
 *     Cisco Technology, Inc - Initial implementation
 *******************************************************************************/

package org.eclipseclp.saros.development.builders;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Map;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.text.Document;
import org.eclipseclp.saros.runtime.ECLiPSeCompilationError;
import org.eclipseclp.saros.runtime.ECLiPSeEngineBuilder;

public class SarosBuilder extends IncrementalProjectBuilder {

	class SourceFileDeltaVisitor implements IResourceDeltaVisitor {
		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.core.resources.IResourceDeltaVisitor#visit(org.eclipse.core.resources.IResourceDelta)
		 */
		public boolean visit(IResourceDelta delta) throws CoreException {
			IResource resource = delta.getResource();
			switch (delta.getKind()) {
			case IResourceDelta.ADDED:
				// handle added resource
				compileSourceFile(resource);
				break;
			case IResourceDelta.REMOVED:
				// handle removed resource
				break;
			case IResourceDelta.CHANGED:
				// handle changed resource
				compileSourceFile(resource);
				break;
			}
			// return true to continue visiting children.
			return true;
		}
	}

	class SourceFileResourceVisitor implements IResourceVisitor {
		public boolean visit(IResource resource) {
			compileSourceFile(resource);
			// return true to continue visiting children.
			return true;
		}
	}

	public static final String BUILDER_ID = "org.eclipseclp.saros.development.SarosBuilder";

	private static final String MARKER_TYPE = "org.eclipseclp.saros.development.SarosProblem";

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.core.internal.events.InternalBuilder#build(int,
	 *      java.util.Map, org.eclipse.core.runtime.IProgressMonitor)
	 */
	protected IProject[] build(int kind, Map args, IProgressMonitor monitor)
			throws CoreException {
		if (kind == FULL_BUILD) {
			fullBuild(monitor);
		} else {
			IResourceDelta delta = getDelta(getProject());
			if (delta == null) {
				fullBuild(monitor);
			} else {
				incrementalBuild(delta, monitor);
			}
		}
		return null;
	}

	void compileSourceFile(IResource resource) {
		if (resource instanceof IFile
				&& (resource.getName().endsWith(".ecl") || resource.getName()
						.endsWith(".pl"))) {
			IFile file = (IFile) resource;
			deleteMarkers(file);
			ECLiPSeEngineBuilder compiler = new ECLiPSeEngineBuilder(resource);
			ArrayList errors = compiler.getCompilationResults();
			ECLiPSeCompilationError error = null;
			IFile errorFile = null;
			for (int count = 0; count < errors.size(); count++) {
				error = (ECLiPSeCompilationError) errors.get(count);
				if (error.getFile() == null) {
					errorFile = file;
					addMarker(errorFile, error.getDetail(), error
							.getLineNumber(), -1, -1, error.getSeverity());
				} else {
					Document document = null;
					/*
					 * ToDo: Get the document from the referenced editors
					 * IWorkbenchWindow[] windows = PlatformUI.getWorkbench()
					 * .getWorkbenchWindows(); for (int window = 0; window <
					 * windows.length; window++) { IWorkbenchPage[] pages =
					 * windows[window].getPages(); for (int page = 0; page <
					 * pages.length; page++) { IEditorReference[]
					 * editorReferences = pages[page] .getEditorReferences();
					 * for (int ref = 0; ref < editorReferences.length; ref++) {
					 * if (editorReferences[ref].getEditor(false) instanceof
					 * ECLiPSeSourceEditor) { if
					 * (editorReferences[ref].isDirty()) { continue; }
					 * ECLiPSeSourceEditor editor = (ECLiPSeSourceEditor)
					 * editorReferences[ref] .getEditor(false); System.err
					 * .println("None dirty editor: " + editor.getEditorInput()
					 * .getName()); } } } }
					 */
					errorFile = file.getProject().getFile(error.getFile());
					String errMsg = error.getType() + ": " + error.getReason();
					int lineNumber = error.getLineNumber();
					int startChar = error.getStartChar();
					int endChar = error.getEndChar();
					if (startChar != -1 && endChar != -1) {
						try {
							InputStream is = errorFile.getContents();
							byte[] contents = new byte[is.available()];
							is.read(contents);
							document = new Document(new String(contents));
							int lineStartCharPos;
							lineStartCharPos = document
									.getLineOffset(lineNumber - 1);
							startChar += lineStartCharPos;
							endChar += lineStartCharPos;
						} catch (Exception e) {
							startChar = -1;
							endChar = -1;
							errMsg = errMsg + "\n" + error.getDetail();
						}
					}

					addMarker(errorFile, errMsg, lineNumber, startChar,
							endChar, error.getSeverity());
				}
			}
		}
	}

	private void deleteMarkers(IFile file) {
		try {
			file.deleteMarkers(MARKER_TYPE, false, IResource.DEPTH_ZERO);
		} catch (CoreException ce) {
		}
	}

	private void addMarker(IFile file, String message, int lineNumber,
			int charStart, int charEnd, int severity) {
		try {
			IMarker marker = file.createMarker(MARKER_TYPE);
			marker.setAttribute(IMarker.MESSAGE, message);
			marker.setAttribute(IMarker.SEVERITY, severity);
			if (lineNumber == -1) {
				lineNumber = 1;
			} else if (charStart != -1 && charEnd != -1) {
				marker.setAttribute(IMarker.CHAR_START, charStart);
				marker.setAttribute(IMarker.CHAR_END, charEnd);
			}
			marker.setAttribute(IMarker.LINE_NUMBER, lineNumber);
		} catch (CoreException e) {
		}
	}

	protected void fullBuild(final IProgressMonitor monitor)
			throws CoreException {
		try {
			getProject().accept(new SourceFileResourceVisitor());
		} catch (CoreException e) {
		}
	}

	protected void incrementalBuild(IResourceDelta delta,
			IProgressMonitor monitor) throws CoreException {
		delta.accept(new SourceFileDeltaVisitor());
	}
}
