/*******************************************************************************
 * Copyright (c) 2006 Cisco Technology, Inc and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Author: Andrew Cheadle, CrossCore Optimization Limited
 *
 * Contributors:
 *     Cisco Technology, Inc - Initial implementation
 *******************************************************************************/

package org.eclipseclp.saros.development.wizards;

import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.ui.INewWizard;
import org.eclipse.ui.IWorkbench;
import org.eclipse.core.runtime.*;
import org.eclipse.jface.operation.*;
import java.lang.reflect.InvocationTargetException;

import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.CoreException;
import java.io.*;
import org.eclipse.ui.*;
import org.eclipse.ui.ide.IDE;
import org.eclipseclp.saros.core.preferences.LaunchQueryComposite;
import org.eclipseclp.saros.core.preferences.PreferenceConstants;
import org.eclipseclp.saros.development.Saros;

public class NewModuleWizard extends Wizard implements INewWizard {
	private NewModuleWizardPage page;

	private ISelection selection;

	/**
	 * Constructor for NewSourceFileWizard.
	 */
	public NewModuleWizard() {
		super();
		setNeedsProgressMonitor(true);
	}

	/**
	 * Adding the page to the wizard.
	 */

	public void addPages() {
		page = new NewModuleWizardPage(selection);
		addPage(page);
	}

	/**
	 * This method is called when 'Finish' button is pressed in the wizard. We
	 * will create an operation and run it using wizard as execution context.
	 */
	public boolean performFinish() {
		final String containerName = page.getContainerName();
		final String moduleName = page.getModuleName();
		final LaunchQueryComposite queryGroup = page.getLaunchQueryComposite();
		queryGroup.commit();
		IRunnableWithProgress op = new IRunnableWithProgress() {
			public void run(IProgressMonitor monitor)
					throws InvocationTargetException {
				try {
					doFinish(containerName, moduleName, queryGroup, monitor);
				} catch (CoreException e) {
					throw new InvocationTargetException(e);
				} finally {
					monitor.done();
				}
			}
		};
		try {
			getContainer().run(true, false, op);
		} catch (InterruptedException e) {
			return false;
		} catch (InvocationTargetException e) {
			Throwable realException = e.getTargetException();
			MessageDialog.openError(getShell(), "Error", realException
					.getMessage());
			return false;
		}
		return true;
	}

	/**
	 * The worker method. It will find the container, create the file if missing
	 * or just replace its contents, and open the editor on the newly created
	 * file.
	 */

	private void doFinish(final String containerName, final String moduleName,
			final LaunchQueryComposite queryGroup,
			final IProgressMonitor monitor) throws CoreException {
		monitor.beginTask("Creating " + moduleName, 3);
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		IResource resource = root.findMember(new Path(containerName));
		if (!resource.exists() || !(resource instanceof IContainer)) {
			throwCoreException("Container \"" + containerName
					+ "\" does not exist.");
		}
		// Note won't have got this far if the file already exists
		monitor.setTaskName("Setting ECLiPSe module '" + moduleName
				+ "' properties...");
		IContainer container = (IContainer) resource;
		final IFile file = container.getFile(new Path(moduleName + "."
				+ PreferenceConstants.EclExtension));
		try {
			InputStream stream = openContentStream(moduleName);
			if (file.exists()) {
				file.setContents(stream, true, true, monitor);
			} else {
				file.create(stream, true, monitor);
			}
			stream.close();
		} catch (IOException e) {
		}
		monitor.worked(1);
		getShell().getDisplay().asyncExec(new Runnable() {
			public void run() {
				try {
					file.setPersistentProperty(new QualifiedName(
							PreferenceConstants.SarosPlugin,
							PreferenceConstants.FileLaunchQuery), ""
							+ queryGroup.isFileSpecificLaunchQuery());
					file.setPersistentProperty(new QualifiedName(
							PreferenceConstants.SarosPlugin,
							PreferenceConstants.HasLaunchQuery), ""
							+ queryGroup.hasLaunchQuery());
					file.setPersistentProperty(new QualifiedName(
							PreferenceConstants.SarosPlugin,
							PreferenceConstants.LaunchFile), queryGroup
							.getFileName());
					file.setPersistentProperty(new QualifiedName(
							PreferenceConstants.SarosPlugin,
							PreferenceConstants.LaunchModule), queryGroup
							.getModuleName());
					file.setPersistentProperty(new QualifiedName(
							PreferenceConstants.SarosPlugin,
							PreferenceConstants.LaunchQuery), queryGroup
							.getQuery());
					file.setPersistentProperty(new QualifiedName(
							PreferenceConstants.SarosPlugin,
							PreferenceConstants.LaunchRunMode), queryGroup
							.getRunMethod());
				} catch (CoreException e) {
					Status status = new Status(
							IStatus.ERROR,
							"Saros",
							0,
							"Exception caught whilst setting project properties: ",
							e);
					ErrorDialog.openError(Saros.getDefault().getWorkbench()
							.getDisplay().getActiveShell(),
							"Saros Development Plugin Error", e.getMessage(),
							status);
				}
				monitor.worked(1);
				monitor.setTaskName("Opening ECLiPSe module '" + moduleName
						+ "' for editing...");
				IWorkbenchPage page = PlatformUI.getWorkbench()
						.getActiveWorkbenchWindow().getActivePage();
				try {
					IDE.openEditor(page, file, true);
				} catch (PartInitException e) {
				}
			}
		});
		monitor.worked(1);
	}

	/**
	 * We will initialize file contents with a module decalaration
	 */

	private InputStream openContentStream(String moduleName) {
		String contents = ":- module(" + moduleName + ").";
		return new ByteArrayInputStream(contents.getBytes());
	}

	private void throwCoreException(String message) throws CoreException {
		IStatus status = new Status(IStatus.ERROR,
				"org.eclipseclp.saros.development", IStatus.OK, message, null);
		throw new CoreException(status);
	}

	/**
	 * We will accept the selection in the workbench to see if we can initialize
	 * from it.
	 * 
	 * @see IWorkbenchWizard#init(IWorkbench, IStructuredSelection)
	 */
	public void init(IWorkbench workbench, IStructuredSelection selection) {
		this.selection = selection;
	}
}
