/*******************************************************************************
 * Copyright (c) 2006 Cisco Technology, Inc and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Author: Andrew Cheadle, CrossCore Optimization Limited
 *
 * Contributors:
 *     Cisco Technology, Inc - Initial implementation
 *******************************************************************************/

package org.eclipseclp.saros.development.wizards;

import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbench;
import org.eclipse.core.runtime.*;
import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.ui.*;
import org.eclipse.ui.actions.WorkspaceModifyOperation;
import org.eclipse.ui.wizards.newresource.BasicNewProjectResourceWizard;
import org.eclipseclp.saros.core.preferences.PreferenceConstants;
import org.eclipseclp.saros.core.preferences.PreferenceMessages;
import org.eclipseclp.saros.development.Saros;
import org.eclipseclp.saros.development.builders.SarosBuilder;
import org.eclipseclp.saros.development.builders.SarosNature;

public class NewProjectWizard extends BasicNewProjectResourceWizard {
	public NewProjectWizard() {
		super();
		setNeedsProgressMonitor(true);
	}

	/**
	 * Adding the page to the wizard.
	 */

	public void addPages() {
		projectPage = new NewProjectWizardProjectPage(selection);
		addPage(projectPage);
		enginePage = new NewProjectWizardEnginePage(selection);
		addPage(enginePage);
	}

	/**
	 * This method is called when 'Finish' button is pressed in the wizard. We
	 * will create an operation and run it using wizard as execution context.
	 */
	public boolean performFinish() {
		final NewProjectWizardProjectComposite projectWizard = projectPage
				.getWizardComposite();
		final NewProjectWizardEngineComposite engineWizard = enginePage
				.getWizardComposite();
		projectWizard.commit(); // Must be done in GUI thread
		engineWizard.commit(); // Must be done in GUI thread

		WorkspaceModifyOperation op = new WorkspaceModifyOperation() {
			public void execute(IProgressMonitor monitor)
					throws InvocationTargetException {
				try {
					createProject(projectWizard, engineWizard, monitor);
				} catch (CoreException e) {
					throw new InvocationTargetException(e);
				} finally {
					monitor.done();
				}
			}
		};
		try {
			getContainer().run(true, false, op);
		} catch (InterruptedException e) {
			return false;
		} catch (InvocationTargetException e) {
			Throwable realException = e.getTargetException();
			MessageDialog.openError(getShell(), "Error", realException
					.getMessage());
			return false;
		}
		return true;
	}

	private void createProject(NewProjectWizardProjectComposite projectWizard,
			NewProjectWizardEngineComposite engineWizard,
			IProgressMonitor monitor) throws CoreException {
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		Path projectPath = new Path(projectWizard.getProjectPath());
		monitor.subTask(PreferenceMessages
				.getString("ECLiPSeStrings.CreatingProject"));
		IProject project = root.getProject(projectWizard.getProjectName());
		IProjectDescription description = ResourcesPlugin.getWorkspace()
				.newProjectDescription(project.getName());
		boolean externalProject = false;
		if (!Platform.getLocation().equals(projectPath)) {
			description.setLocation(projectPath);
			externalProject = true;
		}
		monitor.worked(10);
		description.setNatureIds(new String[] { SarosNature.NATURE_ID });
		ICommand command = description.newCommand();
		command.setBuilderName(SarosBuilder.BUILDER_ID);
		description.setBuildSpec(new ICommand[] { command });
		project.create(description, monitor);
		project.open(monitor);
		monitor.worked(10);
		IPath projectIPath = project.getFullPath();
		String[] outputFolders = null;
		if (externalProject) {
			createLinkedResources(projectPath.toFile(), "", projectIPath,
					monitor);
			// ToDo make a decision!
			outputFolders = new String[0];
		} else {
			outputFolders = projectWizard.getOutputFolders();
		}
		boolean[] compilationTargets = engineWizard
				.getOutputCompilationTargets();
		try {
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.ProjectFolderOutput), ""
					+ projectWizard.isProjectFolderOutput());
			if (outputFolders.length == 0) {
				project.setPersistentProperty(new QualifiedName(
						PreferenceConstants.SarosPlugin,
						PreferenceConstants.ECLiPSeSourceFolderName), "");
				project.setPersistentProperty(new QualifiedName(
						PreferenceConstants.SarosPlugin,
						PreferenceConstants.ECLiPSeObjectFolderName), "");
				project.setPersistentProperty(new QualifiedName(
						PreferenceConstants.SarosPlugin,
						PreferenceConstants.ECLiPSeDocFolderName), "");
				project.setPersistentProperty(new QualifiedName(
						PreferenceConstants.SarosPlugin,
						PreferenceConstants.ECLiPSeHtmlFolderName), "");
			} else {
				IFolder folder = root.getFolder(projectIPath
						.append(outputFolders[0]));
				createFolderHelper(folder, monitor);
				project.setPersistentProperty(new QualifiedName(
						PreferenceConstants.SarosPlugin,
						PreferenceConstants.ECLiPSeSourceFolderName),
						outputFolders[0]);
				folder = root.getFolder(projectIPath.append(outputFolders[1]));
				createFolderHelper(folder, monitor);
				project.setPersistentProperty(new QualifiedName(
						PreferenceConstants.SarosPlugin,
						PreferenceConstants.ECLiPSeObjectFolderName),
						outputFolders[1]);
				folder = root.getFolder(projectIPath.append(outputFolders[2]));
				createFolderHelper(folder, monitor);
				project.setPersistentProperty(new QualifiedName(
						PreferenceConstants.SarosPlugin,
						PreferenceConstants.ECLiPSeDocFolderName),
						outputFolders[2]);
				folder = root.getFolder(projectIPath.append(outputFolders[3]));
				createFolderHelper(folder, monitor);
				project.setPersistentProperty(new QualifiedName(
						PreferenceConstants.SarosPlugin,
						PreferenceConstants.ECLiPSeHtmlFolderName),
						outputFolders[3]);
			}
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.IsProjectSpecificInstallation), ""
					+ engineWizard.isProjectSpecificInstallation());
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.ECLiPSeInstallationName), engineWizard
					.getInstallationName());

			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.IsProjectSpecificCompilationTarget), ""
					+ engineWizard.isProjectSpecificCompilationTarget());
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.ECLiPSeFcompileTarget), ""
					+ compilationTargets[0]);
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.ECLiPSeIcompileTarget), ""
					+ compilationTargets[1]);
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.ECLiPSeEciToHtml), ""
					+ compilationTargets[2]);

			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.FileLaunchQuery),
					PreferenceConstants.False);
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.HasLaunchQuery), ""
					+ engineWizard.hasProjectLaunchQuery());
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.LaunchFile), engineWizard
					.getLaunchFileName());
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.LaunchModule), engineWizard
					.getLaunchModuleName());
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.LaunchQuery), engineWizard
					.getLaunchQuery());
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.LaunchRunMode), engineWizard
					.getLaunchRunMethod());

			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.IsProjectSpecificEngineSettings),
					"false");
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.IsProjectSpecificOutputOptions),
					"false");
			project.setPersistentProperty(new QualifiedName(
					PreferenceConstants.SarosPlugin,
					PreferenceConstants.IsProjectSpecificGCsettings), "false");
		} catch (CoreException e) {
			Status status = new Status(IStatus.ERROR, "Saros", 0,
					"Exception caught whilst setting project properties: ", e);
			ErrorDialog.openError(Saros.getDefault().getWorkbench()
					.getDisplay().getActiveShell(),
					"Saros Development Plugin Error", e.getMessage(), status);
		} finally {
			monitor.done();
		}
	}

	private void createLinkedResources(File directory, final String folderPath,
			final IPath projectIPath, final IProgressMonitor monitor) {
		final String[] names = directory.list();
		List subDirs = new ArrayList();

		for (int i = 0; i < names.length; i++) {
			final File file = new File(directory, names[i]);
			final boolean isDirectory = file.isDirectory();
			if (isDirectory) {
				if (monitor.isCanceled()) {
					return;
				}
				subDirs.add(file);
			}
			final int j = i;
			Runnable fcr = new Runnable() {
				public void run() {
					try {
						if (isDirectory) {
							IFolder folder = ResourcesPlugin.getWorkspace()
									.getRoot().getFolder(
											projectIPath.append(folderPath)
													.append(names[j]));
							folder.createLink(
									new Path(file.getCanonicalPath()),
									IResource.ALLOW_MISSING_LOCAL
											| IResource.REPLACE, monitor);
						} else {
							IFile ifile = ResourcesPlugin.getWorkspace()
									.getRoot().getFile(
											projectIPath.append(folderPath)
													.append(names[j]));
							ifile.createLink(new Path(file.getCanonicalPath()),
									IResource.ALLOW_MISSING_LOCAL
											| IResource.REPLACE, monitor);
						}
					} catch (Exception e) {
						Status status = new Status(
								IStatus.ERROR,
								"Saros",
								0,
								"Exception caught whilst linking project folders: ",
								e);
						ErrorDialog.openError(Saros.getDefault().getWorkbench()
								.getDisplay().getActiveShell(),
								"Saros Development Plugin Error", e
										.getMessage(), status);
					}
				}
			};
			Display.getDefault().syncExec(fcr);
		}

		while (!subDirs.isEmpty()) {
			File subDir = (File) subDirs.remove(0);
			createLinkedResources(subDir, folderPath + "/" + subDir.getName(),
					projectIPath, monitor);
			if (monitor.isCanceled()) {
				return;
			}
		}
	}

	private void createFolderHelper(final IFolder folder,
			final IProgressMonitor monitor) {
		if (!folder.exists()) {
			IContainer parent = folder.getParent();

			if (parent instanceof IFolder && (!((IFolder) parent).exists())) {

				createFolderHelper((IFolder) parent, monitor);
			}
			Runnable fcr = new Runnable() {
				public void run() {
					try {
						folder.create(false, true, monitor);
					} catch (CoreException e) {
						Status status = new Status(
								IStatus.ERROR,
								"Saros",
								0,
								"Exception caught whilst creating project folders: ",
								e);
						ErrorDialog.openError(Saros.getDefault().getWorkbench()
								.getDisplay().getActiveShell(),
								"Saros Development Plugin Error", e
										.getMessage(), status);
					}
				}
			};
			Display.getDefault().syncExec(fcr);
		}
	}

	/**
	 * We will accept the selection in the workbench to see if we can initialize
	 * from it.
	 * 
	 * @see IWorkbenchWizard#init(IWorkbench, IStructuredSelection)
	 */
	public void init(IWorkbench workbench, IStructuredSelection selection) {
		this.selection = selection;
	}

	private NewProjectWizardProjectPage projectPage = null;

	private NewProjectWizardEnginePage enginePage = null;

	private ISelection selection;
}
