/*******************************************************************************
 * Copyright (c) 2006 Cisco Technology, Inc and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Author: Andrew Cheadle, CrossCore Optimization Limited
 *
 * Contributors:
 *     Cisco Technology, Inc - Initial implementation
 *******************************************************************************/

package org.eclipseclp.saros.runtime;

import com.parctechnologies.eclipse.*;

import java.io.DataOutputStream;
import java.io.IOException;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.custom.VerifyKeyListener;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.widgets.Display;
import org.eclipseclp.saros.core.preferences.PreferenceConstants;

public class ECLiPSeConsoleEngineRuntime extends Thread implements
		QueueListener {

	public ECLiPSeConsoleEngineRuntime(boolean topLevel,
			Environment environment, StyledText console) {
		this.topLevel = topLevel;
		this.environment = environment;
		this.console = console;
		eclipseEngine = environment.startEngine(true);
		if (eclipseEngine != null) {
			start();
		}
	}

	public void run() {
		initialiseQueues();
		try {
			environment.compile();
			eclipseEngine.rpc("set_event_handler(halt, halt/0)");
			if (!topLevel) {
				environment.launchQuery();
			} else {
				Runnable r = new Runnable() {
					public void run() {
						console.addVerifyKeyListener(new VerifyKeyListener() {
							public void verifyKey(VerifyEvent event) {
								totalChars = console.getCharCount();
								engineChars = Math.max(stdoutStreamListener
										.getCharCount(), stderrStreamListener
										.getCharCount());
								if (console.getLineAtOffset(console
										.getCaretOffset()) != console
										.getLineCount()) {
									console.setCaretOffset(totalChars);
								}
								if (event.keyCode == 13) {
									if (console.getText(totalChars - 1,
											totalChars - 1).equals(".")) {
										if (totalChars > engineChars) {
											userInput = console
													.getText(engineChars,
															totalChars - 1);
											notifyEngine();
										}
									}
								} else if (console.getCaretOffset() == engineChars
										&& (event.keyCode == 127 || event.keyCode == 8)) {
									event.doit = false;
								}
							}
						});
					}
				};
				Display.getDefault().syncExec(r);
				eclipseEngine.rpc("sepia_kernel:standalone_toplevel");
			}
		} catch (EclipseException e) {
			if (!terminateRequested) {
				System.err.println("Exception caught in Saros ECLiPSe "
						+ "Console engine: " + e.getMessage());
				e.printStackTrace();
			}
		} catch (IOException e) {
			if (!terminateRequested) {
				System.err.println("Exception caught in Saros ECLiPSe "
						+ "Console: " + e.getMessage());
				e.printStackTrace();
			}
		} finally {
			if (environment.getEngineType() == PreferenceConstants.outOfProcess) {
				try {
					((OutOfProcessEclipse) eclipseEngine).destroy();
				} catch (IOException e) {
				}
			}
		}
	}

	private void initialiseQueues() {
		try {
			stdinOutputQueue = eclipseEngine.getEclipseStdin();
			stdinOutputStream = new DataOutputStream(
					(ToEclipseQueue) stdinOutputQueue);
			stdinOutputQueue.setListener(this);
			stdoutStreamListener = new ECLiPSeStdStreamDataListener(console,
					SWT.COLOR_BLUE, false, false);
			eclipseEngine.getEclipseStdout().setListener(stdoutStreamListener);
			stderrStreamListener = new ECLiPSeStdStreamDataListener(console,
					SWT.COLOR_RED, false, false);
			eclipseEngine.getEclipseStderr().setListener(stderrStreamListener);
			eclipseEngine.rpc(new CompoundTermImpl(",", new CompoundTermImpl(
					"set_stream_property", new Atom("output"),
					new Atom("flush"), new Atom("end_of_line")),
					new CompoundTermImpl("set_stream_property", new Atom(
							"error"), new Atom("flush"),
							new Atom("end_of_line"))));
			String eventQueueName = "event_queue";
			eclipseEngine.rpc("event_create(post_events_from_stream("
					+ eventQueueName + "), [defers], Event)@sepia_kernel,"
					+ "peer_queue_create(" + eventQueueName + ", '"
					+ eclipseEngine.getPeerName().functor() + "',"
					+ "async, bidirect, Event)");
			asyncRPCTerminateQueue = eclipseEngine
					.getAsyncEclipseQueue(eventQueueName);
			asyncRPCTerminateStream = new EXDROutputStream(
					asyncRPCTerminateQueue.getOutputStream());
		} catch (Exception e) {
			System.err.println("Exception caught whilst initialising "
					+ "Saros ECLiPSe Console engine queues: " + e.getMessage());
			e.printStackTrace();
		}
	}

	public void dataAvailable(Object source) {
	}

	synchronized private void notifyEngine() {
		submittedInput = true;
		notifyAll();
	}

	synchronized public void dataRequest(Object source) {
		if (source == stdinOutputQueue) {
			try {
				if (topLevel) {
					while (true) {
						try {
							if (submittedInput) {
								if (totalChars > engineChars) {
									stdinOutputStream.writeBytes(userInput);
									stdinOutputStream.flush();
								}
								submittedInput = false;
								return;
							}
							wait();
						} catch (InterruptedException e) {
						}
					}
				} else {
					RunnableInputDialog inputDialog = new RunnableInputDialog();
					Display.getDefault().syncExec(inputDialog);
					stdinOutputStream.writeBytes(inputDialog.getInput());
					stdinOutputStream.flush();
				}
			} catch (IOException ioe) {
				System.err.println("Exception caught whilst writing input "
						+ "to Saros ECLiPSe Console engine: "
						+ ioe.getMessage());
				ioe.printStackTrace();
			}
		}
	}

	public void terminate() {
		if (terminateRequested) {
			return;
		}
		terminateRequested = true;
		try {
			asyncRPCTerminateStream.write(new Atom("halt"));
			asyncRPCTerminateStream.flush();
		} catch (IOException e) {
		}
	}

	private EclipseEngine eclipseEngine = null;

	private boolean terminateRequested = false;

	private Environment environment = null;

	private boolean topLevel = false;

	StyledText console = null;

	private ToEclipseQueue stdinOutputQueue = null;

	private DataOutputStream stdinOutputStream = null;

	private AsyncEclipseQueue asyncRPCTerminateQueue = null;

	private EXDROutputStream asyncRPCTerminateStream = null;

	private ECLiPSeStdStreamDataListener stdoutStreamListener = null;

	private ECLiPSeStdStreamDataListener stderrStreamListener = null;

	private int totalChars = 0;

	private int engineChars = 0;

	private String userInput = null;

	private boolean submittedInput = false;
}
