/*******************************************************************************
 * Copyright (c) 2006 Cisco Technology, Inc and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Author: Andrew Cheadle, CrossCore Optimization Limited
 *
 * Contributors:
 *     Cisco Technology, Inc - Initial implementation
 *******************************************************************************/

package org.eclipseclp.saros.runtime;

import com.parctechnologies.eclipse.*;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipseclp.saros.core.preferences.PreferenceConstants;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;

public class ECLiPSeEngineBuilder implements QueueListener {

	public ECLiPSeEngineBuilder(IResource sourceFile) {
		errors = new ArrayList();
		this.sourceFile = sourceFile;
		project = sourceFile.getProject();
		eclipseEngine = initialiseEngine();
		if (eclipseEngine != null) {
			initialiseECLiPSeQueues();
			executeBuildCommand();
		}
	}

	public EclipseEngine initialiseEngine() {
		environment = new Environment(Environment.PROJECT_PROPERTIES, project,
				sourceFile);
		return environment.startEngine(true);
	}

	private void initialiseECLiPSeQueues() {
		try {
			stdinOutputQueue = eclipseEngine.getEclipseStdin();
			stdinOutputQueue.setListener(this);
			stdoutInputQueue = eclipseEngine.getEclipseStdout();
			stdoutInputQueue.setListener(this);
			stdoutInputStream = new BufferedInputStream(
					(InputStream) stdoutInputQueue);
			stderrInputQueue = eclipseEngine.getEclipseStderr();
			stderrInputQueue.setListener(this);
			stderrInputStream = new BufferedInputStream(
					(InputStream) stderrInputQueue);
		} catch (EclipseTerminatedException e) {
			System.err.println("Exception caught in Saros ECLiPSe "
					+ "compiler: " + e.getMessage());
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println("Exception caught in Saros ECLiPSe "
					+ "compiler: " + e.getMessage());
			e.printStackTrace();
		}

	}

	private void executeBuildCommand() {
		String sourceFileName = sourceFile.getRawLocation().toOSString();
		String projectDir = project.getLocation().toOSString() + File.separator;
		boolean[] compilationTargets = environment.getCompilationTargets();
		final String[] outputFolders = environment.getOutputFolders();

		try {
			compiling = true;
			eclipseEngine.rpc(new CompoundTermImpl(":", new Atom("tracer_tcl"),
					new CompoundTermImpl("saros_compile", sourceFileName)));
			compiling = false;

			try {
				if (compilationTargets[0]) {
					eclipseEngine.rpc(":", new Atom("tracer_tcl"),
							new CompoundTermImpl("saros_fcompile",
									sourceFileName, projectDir
											+ outputFolders[1]));
				}
			} catch (EclipseException e) {
			}

			try {
				if (compilationTargets[1]) {
					eclipseEngine.rpc(":", new Atom("tracer_tcl"),
							new CompoundTermImpl("saros_icompile",
									sourceFileName, projectDir
											+ outputFolders[2]));
				}
			} catch (EclipseException e) {
			}

			try {
				if (compilationTargets[2]) {
					/*
					 * String name = new File(sourceFileName).getName(); name =
					 * name.substring(0, name.lastIndexOf('.')) + ".eci";
					 * eclipseEngine.rpc("tracer_tcl:saros_eci_to_html('" +
					 * projectDir + outputFolders[2] + File.separator + name +
					 * "','" + projectDir + outputFolders[3] + "', \"\")");
					 */
					// ToDo For a large project this would be sluggish - make it
					// an option!
					eclipseEngine.rpc(":", new Atom("tracer_tcl"),
							new CompoundTermImpl("saros_ecis_to_htmls",
									projectDir + outputFolders[2], projectDir
											+ outputFolders[3], "", project
											.getName()));
				}
			} catch (EclipseException e) {
			}

			if (environment.getEngineType() == PreferenceConstants.outOfProcess) {
				((OutOfProcessEclipse) eclipseEngine).destroy();
			}
		} catch (EclipseException e) {
			System.err.println("Exception caught in Saros ECLiPSe "
					+ "compiler: " + e.getMessage());
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println("Exception caught in Saros ECLiPSe "
					+ "compiler: " + e.getMessage());
			e.printStackTrace();
		}

		try {
			project.refreshLocal(IProject.DEPTH_INFINITE, null);
		} catch (CoreException e) {
			Status status = new Status(IStatus.ERROR, "Saros", 0,
					"Exception caught whilst refreshing workspace: ", e);
			ErrorDialog.openError(Saros.getDefault().getWorkbench()
					.getDisplay().getActiveShell(),
					"Saros Development Plugin Error", e.getMessage(), status);
		}
	}

	public void dataAvailable(Object source) {
		try {
			if (source == stdoutInputQueue) {
				int numBytes = stdoutInputStream.available();
				byte[] bytes = new byte[numBytes];
				if (stdoutInputStream.read(bytes, 0, numBytes) != -1
						&& compiling) {
					errors.add(new ECLiPSeCompilationError(new String(bytes),
							IMarker.SEVERITY_WARNING));
				}
			} else { // source == stderrInputStream
				int numBytes = stderrInputStream.available();
				byte[] bytes = new byte[numBytes];
				if (stderrInputStream.read(bytes, 0, numBytes) != -1
						&& compiling) {
					errors.add(new ECLiPSeCompilationError(new String(bytes),
							IMarker.SEVERITY_ERROR, environment
									.getOutputFolders()[0]));
				}
			}
		} catch (IOException ioe) {
			System.err.println("I/O exception caught whilst reading "
					+ "data from Saros ECLiPSe engine: " + ioe.getMessage());
			ioe.printStackTrace();
		}
	}

	public void dataRequest(Object source) {
		System.err.println("Illegal data requested on: " + source);
	}

	public ArrayList getCompilationResults() {
		return errors;
	}

	private EclipseEngine eclipseEngine = null;

	private ToEclipseQueue stdinOutputQueue = null;

	private FromEclipseQueue stdoutInputQueue = null;

	private BufferedInputStream stdoutInputStream = null;

	private FromEclipseQueue stderrInputQueue = null;

	private BufferedInputStream stderrInputStream = null;

	ArrayList errors = null;

	IProject project = null;

	IResource sourceFile = null;

	private Environment environment = null;

	private boolean compiling = false;
}
