/*******************************************************************************
 * Copyright (c) 2006 Cisco Technology, Inc and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Author: Andrew Cheadle, CrossCore Optimization Limited
 *
 * Contributors:
 *     Cisco Technology, Inc - Initial implementation
 *******************************************************************************/

package org.eclipseclp.saros.runtime;

import com.parctechnologies.eclipse.*;

import java.io.IOException;
import java.io.DataOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Vector;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import org.eclipseclp.saros.core.preferences.PreferenceConstants;

public class ECLiPSeInteractiveEngineRuntime extends Thread implements
		ToECLiPSeInteractionInterface, MultitaskListener, QueueListener {

	public ECLiPSeInteractiveEngineRuntime(
			ECLiPSeInteractionInterface eclipseInteraction,
			final Environment environment) throws Exception {
		toplevelModule = "eclipse";
		toplevelStateMonitor = new SynchronizedCounter();
		multitaskingSemaphore = new SynchronizedCounter();
		tracingSemaphore = new SynchronizedCounter();
		processingInterruptingRPCsSemaphore = new SynchronizedCounter();
		rpcExecutingSemaphore = new SynchronizedCounter();
		this.eclipseInteraction = eclipseInteraction;
		this.environment = environment;
		eclipseEngine = environment.startEngine(true);
		if (eclipseEngine == null) {
			throw new Exception(
					"A valid ECLiPSe installation has not yet been configured.");
		}
		eclipseRPCengine = environment.registerMultitask(this);
		if (eclipseRPCengine == null) {
			throw new Exception("Failed to initialise Saros ECLiPSe engine.");
		}
		executionEnvironmentRPCqueue = new Vector();
		initialiseECLiPSeQueues();

		environment.compile();

		start();
	}

	public void run() {
		try {
			eclipseRPCengine
					.rpc(ECLiPSeInteractionInterface.ECLIPSE_TOPLEVEL_QUERY);
		} catch (Exception e) {
			exceptionUnlessTerminating(e,
					"Exception caught in Saros ECLiPSe engine toplevel-loop");
		}
		environment.shutdownEngine();
	}

	private void initialiseECLiPSeQueues() throws Exception {
		eclipseRPCengine.rpc(ECLiPSeInteractionInterface.ECLIPSE_TOPLEVEL_INIT);

		stdinOutputQueue = eclipseEngine.getEclipseStdin();
		stdinOutputStream = new DataOutputStream(
				(ToEclipseQueue) stdinOutputQueue);
		stdinOutputQueue.setListener(this);
		eclipseEngine.getEclipseStdout().setListener(
				eclipseInteraction.getOutputStreamListener());
		eclipseEngine.getEclipseStderr().setListener(
				eclipseInteraction.getErrorStreamListener());

		answerOutputQueue = eclipseEngine
				.getFromEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_ANSWER_OUT_Q);
		answerOutputQueue.setListener(eclipseInteraction
				.getAnswerStreamListener());
		warningOutputQueue = eclipseEngine
				.getFromEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_WARNING_OUT_Q);
		warningOutputQueue.setListener(eclipseInteraction
				.getWarningStreamListener());

		toplevelOutputQueue = eclipseEngine
				.getFromEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_TOP_LEVEL_OUT_Q);
		toplevelOutputStream = new EXDRInputStream(toplevelOutputQueue);
		toplevelOutputQueue.setListener(this);

		guiInterruptRequestQueue = eclipseEngine
				.getFromEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_INT_REQ_Q);
		guiInterruptRequestStream = new EXDRInputStream(
				guiInterruptRequestQueue);
		guiInterruptRequestQueue.setListener(this);

		toplevelInputQueue = eclipseEngine
				.getToEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_TOP_LEVEL_IN_Q);
		toplevelInputQueue.setListener(this);
		toplevelInputStream = new EXDROutputStream(toplevelInputQueue);

		guiPauseRequestQueue = eclipseEngine
				.getAsyncEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_PAUSE_REQ_Q);
		guiPauseRequestOutputStream = new EXDROutputStream(guiPauseRequestQueue
				.getOutputStream());

		debugTracelineQueue = eclipseEngine
				.getFromEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_DEBUG_TRACELINE_Q);
		debugTracelineStream = new EXDRInputStream(debugTracelineQueue);
		debugTracelineQueue.setListener(this);

		debugOutputQueue = eclipseEngine
				.getFromEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_DEBUG_OUTPUT_Q);
		debugOutputQueue.setListener(this);

		statisticsOutputQueue = eclipseEngine
				.getFromEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_STATISTICS_Q);
		statisticsOutputQueue.setListener(eclipseInteraction
				.getStatisticsStreamListener());

		delayedGoalQueue = eclipseEngine
				.getFromEclipseQueue(ECLiPSeInteractionInterface.ECLIPSE_DELAYED_GOAL_Q);
		delayedGoalStream = new EXDRInputStream(delayedGoalQueue);
		delayedGoalQueue.setListener(this);
	}

	private void exceptionUnlessTerminating(final Exception e,
			final String message) {
		if (!terminating) {
			Runnable r = new Runnable() {
				public void run() {
					System.err.println("Exception caught in Saros ECLiPSe "
							+ "execution environment: " + e.getMessage());
					e.printStackTrace();
					Status status = new Status(IStatus.ERROR, "Saros", 0,
							message + ": ", e);
					ErrorDialog.openError(Saros.getDefault().getWorkbench()
							.getDisplay().getActiveShell(),
							"Saros Execution Environment Error",
							e.getMessage(), status);
				}
			};
			Display.getDefault().asyncExec(r);
		}
	}

	/* toplevel_out data consumer */

	private void toplevelOutputDataAvailable() {
		try {
			String toplevelOutputString = toplevelOutputStream.readTerm()
					.toString();
			eclipseInteraction.reflectStoppedStatus();
			synchronized (toplevelStateMonitor) {
				if (toplevelOutputString
						.equals(ECLiPSeInteractionInterface.ECLIPSE_YES_STRING)) {
					eclipseInteraction
							.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_YES);
				} else if (toplevelOutputString
						.equals(ECLiPSeInteractionInterface.ECLIPSE_MORE_STRING)) {
					eclipseInteraction
							.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_MORE);
					eclipseInteraction.reflectMoreStatus();
				} else if (toplevelOutputString
						.equals(ECLiPSeInteractionInterface.ECLIPSE_NO_STRING)) {
					eclipseInteraction
							.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_IDLE);
				} else if (toplevelOutputString
						.equals(ECLiPSeInteractionInterface.ECLIPSE_IDLE_STRING)) {
					eclipseInteraction
							.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_IDLE);
				} else if (toplevelOutputString
						.equals(ECLiPSeInteractionInterface.ECLIPSE_ABORT_STRING)) {
					eclipseInteraction
							.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_IDLE);
				}
			}
		} catch (IOException ioe) {
			exceptionUnlessTerminating(ioe,
					"Exception caught whilst reading toplevel output "
							+ "from Saros ECLiPSe engine");
		}
	}

	/* gui_interrupt_request data consumer */

	private void guiInterruptRequestDataAvailable() {
		try {
			guiInterruptRequestStream.readTerm();
		} catch (IOException ioe) {
			exceptionUnlessTerminating(ioe,
					"Exception caught whilst reading gui_interrupt_request "
							+ "from Saros ECLiPSe engine");
		}
		guiInterruptRequestHandler();
	}

	/* debug_traceline data consumer */

	private void debugTracelineDataAvailable() {
		try {
			Collection term = (Collection) debugTracelineStream.readTerm();
			ECLiPSeTraceLine traceLine = ECLiPSeTraceLine.parse(term);
			eclipseInteraction.updateTraceLogData(traceLine.getLineText());
			eclipseInteraction.updateCallStackData(traceLine);
			eclipseInteraction.traceSource(traceLine);
		} catch (IOException ioe) {
			exceptionUnlessTerminating(ioe,
					"Exception caught whilst reading debug traceline "
							+ "from Saros ECLiPSe engine");
		}
	}

	/* debug_output data consumer */

	private void debugOutputDataAvailable() {
		try {
			int numBytes = debugOutputQueue.available();
			byte[] bytes = new byte[numBytes];
			if (debugOutputQueue.read(bytes, 0, numBytes) != -1) {
				eclipseInteraction.updateTraceLogData(new String(bytes));
			}
		} catch (IOException ioe) {
			exceptionUnlessTerminating(ioe,
					"Exception caught whilst reading debug "
							+ "output from Saros ECLiPSe engine");
		}
	}

	/* gui_dg_info data consumer */

	private void delayedGoalDataAvailable() {
		try {
			CompoundTerm term;
			ArrayList delayedGoals = new ArrayList();
			while (true) {
				term = (CompoundTerm) delayedGoalStream.readTerm();
				if (term.functor().equals("end")) {
					break;
				} else {
					delayedGoals.add(ECLiPSeDelayedGoal.parse(term));
				}
			}
			eclipseInteraction.updateDelayedGoalsData(delayedGoals);
		} catch (IOException ioe) {
			exceptionUnlessTerminating(ioe,
					"Exception caught whilst reading delayed goal "
							+ "info from Saros ECLiPSe engine");
		}
	}

	synchronized public void dataAvailable(Object source) {
		// System.err.println("dataAvailable on source: " + source.toString());
		if (source == toplevelOutputQueue) {
			toplevelOutputDataAvailable();
		} else if (source == guiInterruptRequestQueue) {
			guiInterruptRequestDataAvailable();
		} else if (source == debugTracelineQueue) {
			debugTracelineDataAvailable();
		} else if (source == debugOutputQueue) {
			debugOutputDataAvailable();
		} else if (source == delayedGoalQueue) {
			delayedGoalDataAvailable();
		}
		// System.err.println("Exited dataAvailable on source: " +
		// source.toString());
	}

	/* stdin data producer */

	synchronized public void dataRequest(Object source) {
		// System.err.println("dataRequest on source: " + source.toString());
		if (source == stdinOutputQueue) {
			try {
				stdinOutputStream.writeBytes(eclipseInteraction.requestInput());
				stdinOutputStream.flush();
			} catch (IOException ioe) {
				exceptionUnlessTerminating(ioe,
						"Exception caught whilst writing input "
								+ "to Saros ECLiPSe engine");
			}
		} else if (source == toplevelInputQueue) {
			while (true) {
				synchronized (toplevelStateMonitor) {
					processInterruptingRPCqueue();
					try {
						if (moreRequested) {
							goalPending = false;
							goal = null;
							moreRequested = false;
							Atom controlAtom = new Atom("more");
							toplevelInputStream.write(controlAtom);
							toplevelInputStream.flush();
							eclipseInteraction
									.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_RUNNING);
							return;
						}
					} catch (IOException ioe) {
						exceptionUnlessTerminating(ioe,
								"Exception whilst sending more control atom to Saros ECLiPSe engine");
					}
					if (goalPending) {
						if (eclipseInteraction.getToplevelState() == ECLiPSeInteractionInterface.ECLIPSE_STATE_YES) {
							try {
								Atom controlAtom = new Atom("end");
								toplevelInputStream.write(controlAtom);
								toplevelInputStream.flush();
								eclipseInteraction
										.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_IDLE);
								return;
							} catch (IOException ioe) {
								exceptionUnlessTerminating(ioe,
										"IO Exception whilst sending end control atom to Saros ECLiPSe engine");
							}
						} else if (eclipseInteraction.getToplevelState() == ECLiPSeInteractionInterface.ECLIPSE_STATE_MORE) {
							try {
								Atom controlAtom = new Atom("end");
								toplevelInputStream.write(controlAtom);
								toplevelInputStream.flush();
								eclipseInteraction
										.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_RUNNING);
								return;
							} catch (IOException ioe) {
								exceptionUnlessTerminating(ioe,
										"Exception whilst sending end control atom to Saros ECLiPSe engine");
							}
						}
						reconfigureEngine();
						try {
							toplevelInputStream.write(goal);
							toplevelInputStream.flush();
							eclipseInteraction
									.requestToplevelStateTransition(ECLiPSeInteractionInterface.ECLIPSE_STATE_RUNNING);
						} catch (IOException ioe) {
							exceptionUnlessTerminating(ioe,
									"Exception whilst posting goal to Saros ECLiPSe engine");
						}
						goal = null;
						goalPending = false;
						return;
					}
					if (!eclipseInteraction
							.getMultipleExecutionEnvironmentInstanceSupport()) {
						eclipseInteraction.activate();
					}
					try {
						toplevelStateMonitor.wait();
					} catch (InterruptedException e) {
					}
					if (terminating) {
						try {
							Atom controlAtom = new Atom("exit");
							toplevelInputStream.write(controlAtom);
							toplevelInputStream.flush();
							environment.shutdownEngine();
							return;
						} catch (IOException ioe) {
							exceptionUnlessTerminating(ioe,
									"IO Exception whilst sending exit control atom to Saros ECLiPSe engine");
						}
					}
				}
			}
		}
		// System.err.println("Exited dataRequest on source: " +
		// source.toString());
	}

	private void reconfigureEngine() {
		environment.reconfigureEngine();
		setStatisticsCollecting(false);
		eclipseInteraction.updateTraceLogData(null);
		eclipseInteraction.updateCallStackData(null);
		if (eclipseInteraction.getMultipleExecutionEnvironmentInstanceSupport()) {
			eclipseInteraction.sendEngineReconfigurationNotification();
		} else {
			eclipseInteraction.deactivate();
			if (eclipseInteraction.statisticsReportingEnabled()) {
				eclipseInteraction.setStatisticsCollecting(true);
				rpc(
						new CompoundTermImpl(":", new Atom("tracer_tcl"),
								new CompoundTermImpl(
										"start_report_stats_text_summary",
										eclipseInteraction
												.getStatisticsRefreshRate())),
						"Failed to initiate summary statistics collection.");
			}
		}
		rpc(new CompoundTermImpl("set_flag", new Atom("toplevel_module"),
				new Atom(toplevelModule)),
				"Failed to set toplevel module to: '" + toplevelModule + "'");
		if (tracingSemaphore.getValue() != 0) {
			rpc(new CompoundTermImpl("set_flag", new Atom(
					PreferenceConstants.debugging), new Atom(
					PreferenceConstants.creep)),
					"Failed to initialise engine debug tracing.");
		}
	}

	synchronized public void starting(
			EclipseMultitaskConnection eclipseMultitaskConnection, String type) {
		if (type == ECLiPSeInteractionInterface.ECLIPSE_TRACER_MULTITASK_PHASE) {
			initialiseTracerInteraction();
		}
	}

	synchronized public void ending(
			EclipseMultitaskConnection eclipseMultitaskConnection, String type) {
		if (type == ECLiPSeInteractionInterface.ECLIPSE_TRACER_MULTITASK_PHASE) {
			terminateTracerInteraction();
		}
	}

	public void sendToplevelInCommand(String command, String query) {
		synchronized (toplevelStateMonitor) {
			if (goalPending) {
				Shell shell = new Shell();
				MessageBox messageBox = new MessageBox(shell, SWT.OK
						| SWT.ICON_WARNING);
				messageBox
						.setMessage("Saros ECLiPSe is already executing. Abort the running query or let it complete!");
				messageBox.open();
				shell.dispose();
				return;
			}

			if (command.equals(PreferenceConstants.trace)) {
				tracingSemaphore.set();
				command = PreferenceConstants.call;
			}

			goal = new CompoundTermImpl(command, query);
			goalPending = true;
			toplevelStateMonitor.notifyAll();
		}
	}

	public void requestInterrupt() {
		synchronized (toplevelStateMonitor) {
			if (interruptRequested
					|| eclipseInteraction.getToplevelState() != ECLiPSeInteractionInterface.ECLIPSE_STATE_RUNNING) {
				return;
			}
			interruptRequested = true;

			try {
				guiPauseRequestOutputStream.write(new Atom("int"));
				guiPauseRequestOutputStream.flush();
			} catch (IOException ioe) {
				exceptionUnlessTerminating(ioe,
						"Exception caught whilst sending interrupt request "
								+ "to Saros ECLiPSe engine");
			}
		}
	}

	public void requestMore() {
		synchronized (toplevelStateMonitor) {
			if (eclipseInteraction.getToplevelState() != ECLiPSeInteractionInterface.ECLIPSE_STATE_MORE) {
				return;
			}
			moreRequested = true;
			goal = null;
			goalPending = false;
			toplevelStateMonitor.notifyAll();
		}
	}

	public void shutdown() {
		terminating = true;
		synchronized (toplevelStateMonitor) {
			int state = eclipseInteraction.getToplevelState();
			synchronized (tracingSemaphore) {
				if (tracingSemaphore.getValue() != 0) {
					sendTracerCommand(null, new CompoundTermImpl(":", new Atom(
							"tracer_tcl"), new CompoundTermImpl(
							"set_tracer_command",
							PreferenceConstants.tracerLeapCommand)));
				} else {
					if (state == ECLiPSeInteractionInterface.ECLIPSE_STATE_RUNNING) {
						try {
							guiPauseRequestOutputStream
									.write(new Atom("abort"));
							guiPauseRequestOutputStream.flush();
						} catch (IOException ioe) {
						}
					} else {
						toplevelStateMonitor.notifyAll();
					}
				}
			}
		}
	}

	public void setToplevelModule(String module) {
		toplevelModule = module;
	}

	public String getToplevelModule() {
		return toplevelModule;
	}

	private void updateActiveModuleList(boolean interrupting) {
		CompoundTerm listModules = new CompoundTermImpl(":", new Atom(
				"tracer_tcl"), new CompoundTermImpl("list_modules", null, null));

		if (interrupting) {
			AsynchronousRPCresult r = new AsynchronousRPCresult() {
				public void run() {
					refreshActiveModules(getResult());
				}
			};
			interruptingRpc(listModules,
					"Failed to retrieve current module list.", r, null);
			return;
		}

		try {
			final CompoundTerm result = rpc(listModules);
			Runnable r = new Runnable() {
				public void run() {
					refreshActiveModules(result);
				}
			};
			Display.getDefault().asyncExec(r);
		} catch (Exception e) {
			return;
		}

	}

	private void refreshActiveModules(CompoundTerm result) {
		ArrayList moduleList = new ArrayList(
				(Collection) ((CompoundTerm) result.arg(2)).arg(1));
		String[] modules = new String[moduleList.size()];
		for (int index = 0; index < moduleList.size(); index++) {
			modules[index] = ((Atom) moduleList.get(index)).functor()
					.toString();
		}
		String module = ((Atom) ((CompoundTerm) result.arg(2)).arg(2))
				.functor();
		toplevelModule = module;
		eclipseInteraction.setActiveModules(modules, module);
	}

	public void interruptingRpc(CompoundTerm command, String failMessage) {
		interruptingRpc(command, failMessage, null, null);
	}

	public void interruptingRpc(CompoundTerm command, String failMessage,
			AsynchronousRPCresult successContinuation,
			Runnable failureContinuation) {
		synchronized (processingInterruptingRPCsSemaphore) {
			executionEnvironmentRPCqueue.add(new AsynchronousRPC(command,
					failMessage, successContinuation, failureContinuation));
			if (eclipseInteraction
					.getMultipleExecutionEnvironmentInstanceSupport()) {
				if (processingInterruptingRPCsSemaphore.getValue() != 0
						&& executionEnvironmentRPCqueue.size() > 1) {
					return;
				}
			}
		}

		synchronized (multitaskingSemaphore) {
			if (multitaskingSemaphore.getValue() != 0
					&& tracingSemaphore.getValue() != 0) {
				// Don't do this in the GUI thread, may deadlock the
				// sync update of StyledText areas.
				new Thread() {
					public void run() {
						processInterruptingRPCqueue();
					}
				}.start();
				return;
			}
		}

		synchronized (toplevelStateMonitor) {
			int state = eclipseInteraction.getToplevelState();
			if (state == ECLiPSeInteractionInterface.ECLIPSE_STATE_RUNNING) {
				try {
					guiPauseRequestOutputStream.write(new Atom("int"));
					guiPauseRequestOutputStream.flush();
				} catch (IOException ioe) {
					exceptionUnlessTerminating(ioe,
							"Exception caught whilst requesting "
									+ "Saros ECLiPSe engine asynchronous rpc");
				}
			} else {
				toplevelStateMonitor.notifyAll();
			}
		}
	}

	private void processInterruptingRPCqueue() {
		synchronized (processingInterruptingRPCsSemaphore) {
			processingInterruptingRPCsSemaphore.increment();
			if (processingInterruptingRPCsSemaphore.getValue() != 1) {
				processingInterruptingRPCsSemaphore.decrement();
				return;
			}
		}
		while (!executionEnvironmentRPCqueue.isEmpty()) {
			final AsynchronousRPC arpc = (AsynchronousRPC) executionEnvironmentRPCqueue
					.remove(0);
			CompoundTerm result = null;
			try {
				result = rpc(arpc.getCommand());
			} catch (final Exception e) {
				if (e instanceof RpcExecutingException) {
					executionEnvironmentRPCqueue.add(0, arpc);
					return;
				}
				if (terminating) {
					return;
				}
				System.err
						.println("Exception whilst executing asynchronous RPC: "
								+ e.getMessage());
				e.printStackTrace();
				Runnable r = new Runnable() {
					public void run() {
						String reason = e.getMessage();
						if (reason == null) {
							if (e instanceof EclipseException) {
								reason = "ECLiPSe rpc goal failure: "
										+ arpc.getCommand();
							} else {
								reason = "Unknown.";
							}
						}
						Status status = new Status(IStatus.ERROR, "Saros", 0,
								reason, null);
						ErrorDialog
								.openError(Saros.getDefault().getWorkbench()
										.getDisplay().getActiveShell(),
										"Saros Plugin Error",
										arpc.getMessage(), status);
					}
				};
				Display.getDefault().asyncExec(r);
				Runnable failureContinuation = arpc.getFailureContinuation();
				if (failureContinuation != null) {
					Display.getDefault().asyncExec(failureContinuation);
				}
				continue;
			}
			AsynchronousRPCresult successContinuation = arpc
					.getSuccessContinuation();
			if (successContinuation != null) {
				successContinuation.setResult(result);
				Display.getDefault().asyncExec(successContinuation);
			}
		}
		updateActiveModuleList(false);
		processingInterruptingRPCsSemaphore.decrement();
	}

	public String getInstallationPath() {
		return environment.getInstallation().getInstallationPath();
	}

	private void guiInterruptRequestHandler() {
		// Process any execution environment rpcs
		boolean processInterruptingRPCs = false;
		synchronized (toplevelStateMonitor) {
			if (eclipseInteraction.getToplevelState() == ECLiPSeInteractionInterface.ECLIPSE_STATE_RUNNING) {
				processInterruptingRPCs = true;
			} else {
				System.err.println("Notifying toplevel");
				toplevelStateMonitor.notifyAll();
				return;
			}
		}

		if (processInterruptingRPCs) {
			processInterruptingRPCqueue();
		}

		synchronized (toplevelStateMonitor) {
			if (!interruptRequested
					|| eclipseInteraction.getToplevelState() != ECLiPSeInteractionInterface.ECLIPSE_STATE_RUNNING) {
				// The interrupt was purely for an execution environment rpc or
				// state has changed
				interruptRequested = false;
				return;
			}

			Runnable rdb = new Runnable() {
				public void run() {
					MessageBox messageBox = new MessageBox(PlatformUI
							.getWorkbench().getActiveWorkbenchWindow()
							.getShell(), SWT.YES | SWT.NO | SWT.ICON_QUESTION);
					messageBox
							.setMessage("Saros ECLiPSe execution interrupted. Do you wish to abort the running query?");
					if (messageBox.open() == SWT.YES) {
						try {
							guiPauseRequestOutputStream
									.write(new Atom("abort"));
							guiPauseRequestOutputStream.flush();
						} catch (IOException ioe) {
							exceptionUnlessTerminating(ioe,
									"Exception caught whilst sending abort request "
											+ "to Saros ECLiPSe engine");
						}
					}
				}
			};
			Display.getDefault().asyncExec(rdb);
			interruptRequested = false;
		}
	}

	private boolean commitToMultitasking(String phase) {
		try {
			multitaskingSemaphore.increment();
			environment.multitaskConfirm();
		} catch (EclipseException ee) {
			exceptionUnlessTerminating(ee,
					"Exception caught whilst initiating "
							+ "Saros ECLiPSe engine " + phase
							+ " multitask phase");
			return false;
		} catch (IOException ioe) {
			exceptionUnlessTerminating(ioe,
					"Exception caught whilst initiating "
							+ "Saros ECLiPSe engine " + phase
							+ " multitask phase");
			return false;
		}
		return true;
	}

	private void requestMultitaskingTerminate(String phase) {
		try {
			multitaskingSemaphore.increment();
			environment.multitaskTerminate();
		} catch (EclipseException ee) {
			exceptionUnlessTerminating(ee,
					"Exception caught whilst terminating "
							+ "Saros ECLiPSe engine " + phase
							+ " multitask phase");
		} catch (IOException ioe) {
			exceptionUnlessTerminating(ioe,
					"Exception caught whilst terminating "
							+ "Saros ECLiPSe engine " + phase
							+ " multitask phase");
		}
	}

	private void initialiseTracerInteraction() {
		if (commitToMultitasking("tracer")) {
			if (!eclipseInteraction
					.getMultipleExecutionEnvironmentInstanceSupport()) {
				eclipseInteraction.activate();
			}
			tracingSemaphore.set();
			eclipseInteraction.reflectTracerStatus(true);
		}
	}

	private void terminateTracerInteraction() {
		if (!eclipseInteraction
				.getMultipleExecutionEnvironmentInstanceSupport()) {
			eclipseInteraction.deactivate();
		}
		tracingSemaphore.clear();
		requestMultitaskingTerminate("tracer");
		eclipseInteraction.reflectTracerStatus(false);
	}

	public void sendTracerCommand(CompoundTerm preCommandRpc,
			CompoundTerm command) {
		synchronized (tracingSemaphore) {
			if (tracingSemaphore.getValue() == 0) {
				MessageBox messageBox = new MessageBox(PlatformUI
						.getWorkbench().getActiveWorkbenchWindow().getShell(),
						SWT.OK | SWT.ICON_INFORMATION);
				messageBox
						.setMessage("The tracer is not currently active. Please re-run the query using the 'Trace' option.");
				messageBox.open();
				return;
			}
			if (preCommandRpc != null) {
				rpc(preCommandRpc, "Failed to initialise tracer.");
			}
			rpc(command, "Failed to initialise tracer.");
			terminateTracerInteraction();
		}
	}

	public boolean statisticsCollecting() {
		return collectingStatistics;
	}

	public void setStatisticsCollecting(boolean enabled) {
		collectingStatistics = enabled;
	}

	public CompoundTerm rpc(CompoundTerm command) throws EclipseException,
			IOException {
		CompoundTerm result = null;
		synchronized (multitaskingSemaphore) {
			if (multitaskingSemaphore.getValue() == 0) {
				synchronized (rpcExecutingSemaphore) {
					if (rpcExecutingSemaphore.getValue() != 0) {
						throw new RpcExecutingException();
					}
				}
			}
			rpcExecutingSemaphore.set();
		}
		try {
			result = eclipseRPCengine.rpc(command);
			rpcExecutingSemaphore.clear();
		} catch (EclipseException e) {
			rpcExecutingSemaphore.clear();
			throw e;
		} catch (IOException e) {
			rpcExecutingSemaphore.clear();
			throw e;
		}
		return result;
	}

	public CompoundTerm rpc(CompoundTerm command, String message) {
		CompoundTerm result = null;
		try {
			result = rpc(command);
		} catch (EclipseException e) {
			exceptionUnlessTerminating(e, message);
		} catch (IOException e) {
			exceptionUnlessTerminating(e, message);
		}
		return result;
	}

	private class SynchronizedCounter {
		private int value = 0;

		public synchronized int getValue() {
			return value;
		}

		public synchronized int increment() {
			return ++value;
		}

		public synchronized int decrement() {
			return --value;
		}

		public synchronized void set() {
			value = 1;
		}

		public synchronized void clear() {
			value = 0;
		}
	}

	private class RpcExecutingException extends EclipseException {
	}

	private EclipseEngine eclipseEngine = null;

	private EclipseMultitaskConnection eclipseRPCengine = null;

	private ToEclipseQueue stdinOutputQueue = null;

	private DataOutputStream stdinOutputStream = null;

	private EXDRInputStream toplevelOutputStream = null;

	private FromEclipseQueue toplevelOutputQueue = null;

	private FromEclipseQueue answerOutputQueue = null;

	private FromEclipseQueue warningOutputQueue = null;

	private ToEclipseQueue toplevelInputQueue = null;

	private EXDROutputStream toplevelInputStream = null;

	private EXDRInputStream guiInterruptRequestStream = null;

	private FromEclipseQueue guiInterruptRequestQueue = null;

	private EXDROutputStream guiPauseRequestOutputStream = null;

	private AsyncEclipseQueue guiPauseRequestQueue = null;

	private EXDRInputStream debugTracelineStream = null;

	private FromEclipseQueue debugTracelineQueue = null;

	private FromEclipseQueue debugOutputQueue = null;

	private FromEclipseQueue statisticsOutputQueue = null;

	private EXDRInputStream delayedGoalStream = null;

	private FromEclipseQueue delayedGoalQueue = null;

	private ECLiPSeInteractionInterface eclipseInteraction = null;

	private boolean interruptRequested = false;

	private boolean goalPending = false;

	private Object goal = null;

	private boolean moreRequested = false;

	private String toplevelModule = null;

	private Environment environment = null;

	private Vector executionEnvironmentRPCqueue = null;

	private boolean terminating = false;

	private boolean collectingStatistics = false;

	private SynchronizedCounter toplevelStateMonitor = null;

	private SynchronizedCounter multitaskingSemaphore = null;

	private SynchronizedCounter tracingSemaphore = null;

	private SynchronizedCounter processingInterruptingRPCsSemaphore = null;

	private SynchronizedCounter rpcExecutingSemaphore = null;
}
